"""Slices CLI Entrypoint."""

import json
import logging
from pathlib import Path
from typing import Annotated, Any, Optional

import msgspec
import typer
from authlib.common.errors import AuthlibBaseError
from jwt.exceptions import PyJWTError
from msgspec import ValidationError
from slices_bi_client.config import SlicesBiApiClientsConfig, SlicesCentralApiClientsConfig

from slices_cli.users_info import OtherUsersInfo
from slices_cli.util.version import handle_newer_version_available

from . import console, err_console
from .api import ApiClients
from .auth import auth_app
from .bi import bi_app
from .config import delete_auth_token, load_auth_token, parse_auth_token
from .context import Context, ContextObj
from .experiment import experiment_app
from .project import project_app
from .pubkey import pubkey_app

logging.basicConfig(level=logging.WARNING)

app = typer.Typer(help="Slices CLI", no_args_is_help=True, add_completion=False)


def version_callback(version: bool):
    """Show CLI version."""
    if version:
        from importlib.metadata import version as metadata_version

        console.print(f"Slices CLI v{metadata_version('slices-cli')}")
        raise typer.Exit(0)


def _upgrade_old_config(old_config_dict: dict[str, Any]) -> dict[str, Any]:
    res = {
        "central_hub_services": {
            "experiment_authority": old_config_dict.get("experiment_authority"),
            "project_authority": old_config_dict.get("project_authority"),
            "pubkey_authority": old_config_dict.get("pubkey"),
            "user_authority": old_config_dict.get("user_authority"),
        },
        "bi_infrastructures": old_config_dict.get("bi", {}),
        "bi_default_image": "Ubuntu 24.04.1",
    }
    if old_config_dict.get("orchestrator"):
        res["bi_infrastructures"]["bi-orchestrator"] = old_config_dict.get("orchestrator")
    return res


@app.callback(invoke_without_command=True)
def initialize_api_clients(
    ctx: Context,
    api_config: Annotated[
        Optional[Path], typer.Option(exists=True, dir_okay=False, hidden=True, envvar="SLICES_CUSTOM_CONFIG")
    ] = None,
    _version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            "-v",
            callback=version_callback,
            help="Show CLI version.",
            is_eager=True,
        ),
    ] = None,
    skip_upgrade_check: Annotated[
        Optional[bool],
        typer.Option(
            "--no-upgrade-check", help="Do not check for newer version.", envvar="SLICES_CLI_SKIP_UPGRADE_CHECK"
        ),
    ] = False,
):
    """Initialize API clients."""

    # First let the user know if a new version is available
    if not skip_upgrade_check and console.is_terminal:
        handle_newer_version_available()

    central_config = None
    bi_config = None
    if api_config:
        try:
            with api_config.open() as f:
                config_str = f.read()
            config_dict = json.loads(config_str)
            if config_dict.get("pubkey") or config_dict.get("bi"):
                err_console.print("[bold red]You are using an old version of the API config. Please upgrade. :x:")
                console.print(f"Your file '{api_config}' translated to the new format:")
                new_config_dict = _upgrade_old_config(config_dict)
                console.print_json(data=new_config_dict)
                raise typer.Exit(1)

            if config_dict.get("central_hub_services"):
                central_config = msgspec.json.decode(config_str, type=SlicesCentralApiClientsConfig)

            if config_dict.get("bi_infrastructures"):
                bi_config = msgspec.json.decode(config_str, type=SlicesBiApiClientsConfig)

            if not central_config and not bi_config:
                err_console.print(
                    "[bold red]Your config file contains neither SlicesCentralApiClientsConfig or "
                    "SlicesBiApiClientsConfig. (One or both required.) :x:"
                )
                raise typer.Exit(1)
        except ValidationError as err:
            err_console.print("[bold red]Something is wrong with your api configuration :x:")
            err_console.print("\n", str(err))
            raise typer.Exit(1) from err

    try:
        auth_token = load_auth_token()
        user_info = parse_auth_token(auth_token) if auth_token else None
    except (AuthlibBaseError, PyJWTError):
        err_console.print("[bold red]Something is wrong with your user credentials :x:")
        err_console.print("[red]Please login again.")

        delete_auth_token()

        raise typer.Exit(128) from None
    api_clients = ApiClients(central_config, bi_config)
    ctx.obj = ContextObj(api=api_clients, user_info=user_info, other_users_info=OtherUsersInfo(api_clients))


app.add_typer(auth_app, name="auth", rich_help_panel="Commands with subcommands")
app.add_typer(pubkey_app, name="pubkey", rich_help_panel="Commands with subcommands")
app.add_typer(bi_app, name="bi", rich_help_panel="Commands with subcommands")
app.add_typer(experiment_app, name="experiment", rich_help_panel="Commands with subcommands")
app.add_typer(project_app, name="project", rich_help_panel="Commands with subcommands")


if __name__ == "__main__":
    try:
        app()
    except ValidationError as err:
        err_console.print("[bold red]Something is wrong with your input :x:")
        err_console.print(f"[red]{err}[/]")
        raise typer.Exit(1) from err
