"""Authentication flow."""

import contextlib
import socket
import string
import time
import webbrowser
from concurrent.futures import ThreadPoolExecutor
from secrets import SystemRandom
from typing import TYPE_CHECKING, cast

import typer
from authlib.integrations.requests_client import OAuth2Session, OAuthError
from authlib.oauth2.rfc8628 import DEVICE_CODE_GRANT_TYPE
from rich.console import Console

from .config import settings
from .server import wait_for_authorization_response

if TYPE_CHECKING:
    from .token import OAuthToken

CODE_VERIFIER_CHARS = string.ascii_letters + string.digits + "-._~"
CODE_VERIFIER_RANDOM = SystemRandom()

DEVICE_FLOW_POLLING_FREQUENCY = 2

error_console = Console(stderr=True, style="bold red")


class AuthenticationFlow:
    """Authentication Flow."""

    def __init__(self) -> None:
        self.client = OAuth2Session(
            settings["client_id"],
            settings["client_secret"],
            scope=settings["scope"],
            token_endpoint_auth_method="client_secret_post",  # noqa: S106
        )

    def perform_login_flow(self):
        """Perform OAuth Login flow and return authorization code."""

        console = Console()

        callback_port = self._find_open_port()
        code_verifier = self._generate_code_verifier()

        redirect_uri = f"http://localhost:{callback_port}/callback"
        auth_url, state = self.client.create_authorization_url(
            settings["authorization_url"],
            redirect_uri=redirect_uri,
            code_verifier=code_verifier,
        )

        with ThreadPoolExecutor(max_workers=1, thread_name_prefix="AuthFlowServer") as executor:
            http_server_future = executor.submit(wait_for_authorization_response, "localhost", callback_port)
            console.print(f"Please go to {auth_url} and login to continue.")
            webbrowser.open(auth_url, new=1)

            with console.status("[mangenta]Waiting for authentication response") as status:
                try:
                    authorization_response = http_server_future.result()
                except TimeoutError:
                    error_console.print("Timeout while waiting for authorization token.")
                    raise typer.Exit(1) from None
                else:
                    status.update("Retrieving token")
                    console.clear()
                    token = self.client.fetch_token(
                        settings["token_endpoint_url"],
                        state=state,
                        authorization_response=authorization_response,
                        redirect_uri=redirect_uri,
                    )
                    if token:
                        return cast("OAuthToken", token)
                    error_console.print("Could not exchange authorization code for a token.")
                    raise typer.Exit(1)

    def _generate_code_verifier(self):
        return "".join(CODE_VERIFIER_RANDOM.choice(CODE_VERIFIER_CHARS) for _ in range(128))

    def _find_open_port(self, start_port=None, stop_port=None):
        start = start_port or 8080
        stop = stop_port or start + 100
        for port in range(start, stop):
            with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
                try:
                    sock.bind(("127.0.0.1", port))
                    sock.listen(1)
                except OSError:
                    is_open = False
                else:
                    is_open = True
            if is_open:
                return port
        raise ConnectionError("Could not find an open port")


class DeviceFlow:
    """Device Authentication Flow."""

    def __init__(self) -> None:
        self.client = OAuth2Session(
            settings["client_id"],
            settings["client_secret"],
            scope=settings["scope"],
            token_endpoint_auth_method="client_secret_post",  # noqa: S106
        )

    def perform_login_flow(self):
        """Perform OAuth Login flow and return authorization code."""

        console = Console()
        code_verifier = self._generate_code_verifier()

        auth_url, state = self.client.create_authorization_url(
            settings["authorization_url"],
            code_verifier=code_verifier,
            grant_type="urn:ietf:params:oauth:grant-type:device_code",
        )

        device_code_response = self.client.request(
            "POST",
            settings["device_authorization_url"],
            withhold_token=True,
            data={
                "client_id": settings["client_id"],
                "client_secret": settings["client_secret"],
                "scope": settings["scope"],
            },
        )

        if not device_code_response.ok:
            raise ValueError("Could not retrieve device code.")

        device_code_info = device_code_response.json()

        console.print(f"Please go to {device_code_info['verification_uri_complete']} and login to continue.")

        expires_in = device_code_info["expires_in"]
        with console.status("[mangenta]Waiting for authorization"):
            while expires_in > 0:
                time.sleep(DEVICE_FLOW_POLLING_FREQUENCY)
                try:
                    token = self.client.fetch_token(
                        settings["token_endpoint_url"],
                        state=state,
                        grant_type=DEVICE_CODE_GRANT_TYPE,
                        device_code=device_code_info["device_code"],
                    )
                    if token:
                        return cast("OAuthToken", token)

                except OAuthError as err:
                    if err.error != "authorization_pending":
                        raise ValueError("Exception while waiting for device authorization") from err

                    expires_in -= DEVICE_FLOW_POLLING_FREQUENCY
                else:
                    error_console.print("Could not retrieve authentication details from device login flow.")
                    raise typer.Exit(1)

            error_console.print("The user code has expired. Please start the login flow again.")
            raise typer.Exit(1)

    def _generate_code_verifier(self):
        return "".join(CODE_VERIFIER_RANDOM.choice(CODE_VERIFIER_CHARS) for _ in range(128))

    def _find_open_port(self, start_port=None, stop_port=None):
        start = start_port or 8080
        stop = stop_port or start + 100
        for port in range(start, stop):
            with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
                try:
                    sock.bind(("127.0.0.1", port))
                    sock.listen(1)
                except OSError:
                    is_open = False
                else:
                    is_open = True
            if is_open:
                return port
        raise ConnectionError("Could not find an open port")


class RefreshFlow:
    """Refresh Flow."""

    def __init__(self) -> None:
        self.client = OAuth2Session(
            settings["client_id"],
            settings["client_secret"],
            scope=settings["scope"],
            token_endpoint_auth_method="client_secret_post",  # noqa: S106
        )

    def perform_refresh(self, refresh_token: str):
        """Perform OAuth Login flow and return authorization code."""

        return cast("OAuthToken", self.client.refresh_token(settings["token_endpoint_url"], refresh_token))
